from ait.commons.util.aws_client import Aws
from ait.commons.util.common import format_err
from ait.commons.util.local_state import set_bucket
from ait.commons.util.settings import DEFAULT_PROFILE, DEFAULT_REGION
from ait.commons.util.user_profile import set_profile, profile_exists, get_profile


class CmdConfig:
    """
    both admin and user
    aws resource or client used in command - sts (to check valid credentials).
    """

    def __init__(self, args):
        self.args = args

    def run(self):

        try:
            profile = self.args.profile if self.args.profile else DEFAULT_PROFILE
            if self.args.bucket:
                set_bucket(self.args.bucket)

            if self.args.ACCESS_KEY and self.args.SECRET_KEY:  
                set_profile(profile, DEFAULT_REGION, self.args.ACCESS_KEY, self.args.SECRET_KEY)

            # check new profile
            if profile_exists(profile):
                user_profile = get_profile(profile)
                aws = Aws(user_profile)

                if aws.is_valid_credentials():
                    return True, 'Valid credentials' + ('' if aws.is_user else ' for admin use')
                else:
                    return False, 'Invalid credentials'
            else:
                return False, f'Profile \'{profile}\' not found. Please run config command with your access keys'

        except Exception as e:
            return False, format_err(e, 'config')



import json

from botocore.exceptions import ClientError

from ait.commons.util.aws_client import Aws
from ait.commons.util.bucket_policy import DEFAULT_PERMS, allowDownloadStmt, denyDeleteStmt
from ait.commons.util.common import gen_uuid, format_err


class CmdCreate:
    """
    admin only
    aws resource or client used in command - s3 client (put_object), s3 resource (BucketPolicy)
    """

    def __init__(self, aws: Aws, args):
        self.aws = aws
        self.args = args

    def run(self):
        if not self.aws:
            return False, 'You need configure your profile first'

        if self.aws.is_user:
            return False, 'You don\'t have permission to use this command'

        area_name = self.args.NAME
        perms = self.args.p  # optional str, default 'ux'

        # generate random uuid prefix for area name
        area_id = gen_uuid()

        try:
            s3_client = self.aws.common_session.client('s3')
            # new upload areas to be created with tagging instead of metadata
            s3_client.put_object(Bucket=self.aws.bucket_name, Key=(area_id + '/'), Tagging=f'name={area_name}&perms={perms}')

            if perms == DEFAULT_PERMS:
                pass # default perms as set in user policy (ux) applies - no need for further actions (deny or allow)
            else:
                # get bucket policy
                bucket_policy = self.aws.common_session.resource('s3').BucketPolicy(self.aws.bucket_name)
                try:
                    policy_str = bucket_policy.policy
                except ClientError:
                    policy_str = ''

                if policy_str:
                    policy_json = json.loads(policy_str)
                else:  # no bucket policy
                    policy_json = json.loads('{ "Version": "2012-10-17", "Statement": [] }')

                allow_stmt = None
                deny_stmt = None

                for stmt in policy_json['Statement']:
                    if stmt['Sid'] == 'AllowDownload':
                        allow_stmt = stmt
                    elif stmt['Sid'] == 'DenyDelete':
                        deny_stmt = stmt
                
                if 'd' in perms: # e.g 'ud' or 'udx'
                    # allow download
                    self.update_perms(policy_json, allow_stmt, allowDownloadStmt(), area_id)

                if 'x' not in perms: # e.g. 'u' or 'ud'
                    # deny delete
                    self.update_perms(policy_json, deny_stmt, denyDeleteStmt(), area_id)

                try:
                    bucket_policy.put(Policy=json.dumps(policy_json))
                except ClientError:
                    pass

            return True, 'Created upload area with UUID ' + area_id + ' and name ' + area_name

        except Exception as e:
            return False, format_err(e, 'create')


    def update_perms(self, policy, stmt, template, area):
        if not stmt:
            stmt = template
            policy['Statement'].append(stmt)
        if isinstance(stmt['Resource'], str):
            stmt['Resource'] = [stmt['Resource']] + [f'arn:aws:s3:::{self.aws.bucket_name}/{area}/*']
        elif isinstance(stmt['Resource'], list):
            stmt['Resource'].append(f'arn:aws:s3:::{self.aws.bucket_name}/{area}/*')



import os
import filetype

from ait.commons.util.settings import DIR_SUPPORT, MAX_DIR_DEPTH
from ait.commons.util.common import format_err
from ait.commons.util.file_transfer import FileTransfer, TransferProgress, transfer
from ait.commons.util.local_state import get_selected_area


class CmdUpload:
    """
    admin and user
    aws resource or client used in command - s3 resource (Bucket().upload_file)
    """

    def __init__(self, aws, args):
        self.aws = aws
        self.args = args
        self.files = []

    def run(self):

        selected_area = get_selected_area()

        if not selected_area:
            return False, 'No area selected'

        try:

            # filter out any duplicate path after expansion 
            # . -> curent drectory
            # ~ -> user home directory

            ps = []
            for p in self.args.PATH:
                p = os.path.abspath(p)  # Normalize a pathname by collapsing redundant separators and up-level references so that A//B, A/B/, A/./B and A/foo/../B all become A/B.
                if not p in ps:
                    ps.append(p)

            # create list of files to upload
            fs = []

            max_depth = 1  # default
            if DIR_SUPPORT and self.args.r:
                max_depth = MAX_DIR_DEPTH

            exclude = lambda f: f.startswith('.') or f.startswith('__')

            def get_files(upload_path, curr_path, level):
                if level < max_depth:  # skip files deeper than max depth
                    level += 1
                    for f in os.listdir(curr_path):
                        full_path = os.path.join(curr_path, f)
                        # skip hidden files and dirs
                        if not exclude(f):
                            if os.path.isfile(full_path):
                                f_size = os.path.getsize(full_path)
                                rel_path = full_path.replace(upload_path + ('' if upload_path.endswith('/') else '/'), '')
                                fs.append(FileTransfer(path=full_path, key=rel_path, size=f_size))

                            elif os.path.isdir(full_path):
                                get_files(upload_path, full_path, level)

            for p in ps:
                if os.path.isfile(p):  # explicitly specified files, whether hidden or starts with '__' not skipped
                    f_size = os.path.getsize(p)
                    f_name = os.path.basename(p)
                    fs.append(FileTransfer(path=p, key=f_name, size=f_size))

                elif os.path.isdir(p):  # recursively handle dir upload
                    get_files(p, p, 0)

            def upload(idx):
                try:
                    key = selected_area + fs[idx].key

                    # creating a new session for each file upload/thread, as it's unclear whether they're
                    # thread-safe or not

                    sess = self.aws.new_session()

                    if not self.args.o and self.aws.obj_exists(key):
                        fs[idx].status = 'File exists. Use -o to overwrite.'
                        fs[idx].successful = True
                        fs[idx].complete = True
                    else:
                        res = sess.resource('s3')
                        ftype = filetype.guess(fs[idx].path)
                        # default contentType
                        contentType = 'application/octet-stream'
                        if ftype is not None:
                            contentType = ftype.mime
                        contentType += '; dcp-type=data'
                        
                        # upload_file automatically handles multipart uploads via the S3 Transfer Manager
                        # put_object maps to the low-level S3 API request, it does not handle multipart uploads
                        res.Bucket(self.aws.bucket_name).upload_file(Filename=fs[idx].path, Key=key,
                                                                     Callback=TransferProgress(fs[idx]),
                                                                     ExtraArgs={'ContentType': contentType})

                        # if file size is 0, callback will likely never be called
                        # and complete will not change to True
                        # hack
                        if fs[idx].size == 0:
                            fs[idx].status = 'Empty file.'
                            fs[idx].complete = True
                            fs[idx].successful = True

                except Exception as thread_ex:
                    fs[idx].status = 'Upload failed.'
                    fs[idx].complete = True
                    fs[idx].successful = False

            print('Uploading...')
            transfer(upload, fs)

            self.files = [f for f in fs if f.successful]

            if all([f.successful for f in fs]):
                return True, 'Successful upload.'

            else:
                return False, 'Failed upload.'

        except Exception as e:
            return False, format_err(e, 'upload')



# __main__.py

import argparse
import os
import sys

from ait.commons.util.settings import DEFAULT_PROFILE, DEBUG_MODE, NAME, VERSION, DIR_SUPPORT
from ait.commons.util.common import is_valid_project_name, is_valid_area_name, is_valid_uuid, INGEST_UPLOAD_AREA_PREFIX
from ait.commons.util.cmd import Cmd
from ait.commons.util.bucket_policy import ALLOWED_PERMS, DEFAULT_PERMS


def valid_project_name(string):
    if not is_valid_project_name(string):
        msg = 'invalid - needs to be between 1-36 alphanumeric characters with no space'
        raise argparse.ArgumentTypeError(msg)
    return string


def valid_area(string):
    if not is_valid_area_name(string):
        msg = 'invalid - needs to be <uuid> or <uuid>/'
        raise argparse.ArgumentTypeError(msg)
    return string


def valid_path(path):
    if os.path.exists(path):  # true if the path is a file, directory, or a valid symlink.
        return path
    else:
        raise argparse.ArgumentTypeError(f"'{path}' is not a valid path")


def valid_remote_path(path): 
    # file, dir/file, dir, dir/, dir/dir1, etc
    err_msg = f"'{path}' is not a valid path. e.g. paths - file, dir/file, dir, dir/, dir/dir1, etc"

    if path.startswith('/'):
        raise argparse.ArgumentTypeError(err_msg)
    else:

        if os.path.isabs(f'/{path}'):  # leading / is needed for isabs to return true, BUT it is invalid if provided path starts with /
            #isabs also normalises path e.g remove double //, etc
            return path
        else:
            raise argparse.ArgumentTypeError(err_msg)


def valid_ingest_upload_area(upload_area):
    # expected format: {INGEST_UPLOAD_AREA_PREFIX}-<env>/<uuid>/
    if upload_area.startswith(INGEST_UPLOAD_AREA_PREFIX):
        without_pref = upload_area[len(INGEST_UPLOAD_AREA_PREFIX):]
        parts = without_pref.split('/')
        if len(parts) > 2:
            env_part = parts[0]
            uuid_part = parts[1]
            envs = [
                'dev',
               #'integration',
                'staging',
                'prod'
            ]
            if env_part in envs and is_valid_uuid(uuid_part):
                bucket = INGEST_UPLOAD_AREA_PREFIX.replace('s3://', '') + env_part
                return bucket, env_part, uuid_part

    msg = f'invalid - expected format {INGEST_UPLOAD_AREA_PREFIX}_ENV_/_UUID_/'
    raise argparse.ArgumentTypeError(msg)


def parse_args(args):
    parser = argparse.ArgumentParser(description=NAME)
    parser.add_argument('--version', '-v', action='version', version=f'{NAME} {VERSION}')

    cmd_parser = parser.add_subparsers(title='command', dest='command')

    cmd_parser.required = True

    parser_config = cmd_parser.add_parser('config', help='configure AWS credentials')
    parser_config.add_argument('ACCESS_KEY', help='AWS Access Key ID', nargs='?')
    parser_config.add_argument('SECRET_KEY', help='AWS Secret Access Key', nargs='?')
    parser_config.add_argument('--bucket', help='use BUCKET instead of default bucket')

    parser_create = cmd_parser.add_parser('create', help='create an upload area (authorised users only)')
    parser_create.add_argument('NAME', help='name for the new area', type=valid_project_name)
    parser_create.add_argument('-p', choices=ALLOWED_PERMS, default=DEFAULT_PERMS, help=f'allowed actions ('
                                                                                        f'permissions) on new '
                                                                                        f'area. u for upload, '
                                                                                        f'x for delete and d for '
                                                                                        f'download. Default is '
                                                                                        f'{DEFAULT_PERMS}')

    parser_select = cmd_parser.add_parser('select', help='select or show the active upload area')
    parser_select.add_argument('AREA', help='area uuid', type=valid_area, nargs='?')

    # cmd_parser.add_parser('dir', help='display active (selected) directory')

    # parser_clear = cmd_parser.add_parser('clear', help='clear current selection')
    # parser_clear.add_argument('-a', action='store_true', help='clear all - selection and known dirs')

    parser_list = cmd_parser.add_parser('list', help='list contents of the area')
    parser_list.add_argument('-b', action='store_true', help='list all areas in the S3 bucket (authorised users only)')

    # parser_upload = cmd_parser.add_parser('upload', help='upload files to the area')
    # group_upload = parser_upload.add_mutually_exclusive_group(required=True)

    # group_upload.add_argument('-a', action='store_true', help='upload all files')
    # group_upload.add_argument('-f', metavar='file', nargs='+',
    #                           help='upload specified file(s)', type=argparse.FileType('r'))
    # parser_upload.add_argument('-o', action='store_true', help='overwrite files with same names')

    parser_upload = cmd_parser.add_parser('upload', help='upload files to the area')
    parser_upload.add_argument('PATH', help='valid file or directory', type=valid_path, nargs='+')
    if DIR_SUPPORT:
        parser_upload.add_argument('-r', action='store_true', help='recursively upload sub-directories')
        parser_upload.add_argument('-d', metavar='DIR', help='upload to specified directory')
    parser_upload.add_argument('-o', action='store_true', help='overwrite files with same names')


    parser_download = cmd_parser.add_parser('download', help='download files from the area')
    group_download = parser_download.add_mutually_exclusive_group(required=True)

    group_download.add_argument('-a', action='store_true', help='download all files from selected area')
    group_download.add_argument('-f', metavar='file', nargs='+', help='download specified file(s) only')

    parser_delete = cmd_parser.add_parser('delete', help='delete files from the area')
    parser_delete.add_argument('PATH', help='path to file or directory to delete', type=valid_remote_path, nargs='*')
    group_delete = parser_delete.add_mutually_exclusive_group(required=False)
    group_delete.add_argument('-a', action='store_true', help='delete all files from the area')
    group_delete.add_argument('-d', action='store_true', help='delete upload area and contents (authorised users only)')

    parser_sync = cmd_parser.add_parser('sync', help='copy data from selected upload area to ingest upload area (authorised users only)')
    parser_sync.add_argument('INGEST_UPLOAD_AREA', help='Ingest upload area', type=valid_ingest_upload_area)

    ps = [parser]
    if DEBUG_MODE:
        ps = [parser, parser_config, parser_create, parser_select, parser_list, parser_upload, parser_download, parser_delete, parser_sync]

    for p in ps:
        p.add_argument(
            '--profile',
            help=f'use PROFILE instead of default \'{DEFAULT_PROFILE}\' profile',
            default=DEFAULT_PROFILE
        )

    # parser.add_argument(
    #     '--region',
    #     help=f'use REGION instead of default \'{DEFAULT_REGION}\' region',
    #     default=DEFAULT_REGION
    # )

    return parser.parse_args(args)


def main():
    try:
        parsed_args = parse_args(sys.argv[1:])
        Cmd(parsed_args)
    except KeyboardInterrupt:
        # If SIGINT is triggered whilst threads are active (upload/download) we kill the entire process to give the
        # user an instant exist, rather than have to hammer on ctrl+c multiple times with various obscure messages.
        #
        # However, os_.exit() is nasty because it allows the Python interpreter to do no cleanup
        # One alternative is to make our transfer threads daemon. However, there are then other non-daemon threads
        # employed by boto for the multi-part transfer. Here, configuring boto to only use the main thread for transfer
        # works but with unmeasured effects on transfer speed. Alternatively, it might be possible to call
        # boto3 client.abort_multipart_upload() but don't know for sure and in any case this requires more work to
        # keep hold of upload IDs, etc.
        #
        # So for now risking the kill though might want to revisit this decision.
        os._exit(0)

if __name__ == '__main__':
	main()



import os
import sys
from datetime import date
import requests

from ait.commons.util.aws_client import Aws
from ait.commons.util.command.config import CmdConfig
from ait.commons.util.command.create import CmdCreate
from ait.commons.util.command.delete import CmdDelete
from ait.commons.util.command.download import CmdDownload
from ait.commons.util.command.list import CmdList
from ait.commons.util.command.select import CmdSelect
from ait.commons.util.command.upload import CmdUpload
from ait.commons.util.command.sync import CmdSync
from ait.commons.util.local_state import get_bucket, set_attr, get_attr
from ait.commons.util.user_profile import profile_exists, get_profile
from ait.commons.util.settings import NAME, VERSION


class Cmd:
    """
    steps to perform before executing command
    if cmd is config, skip steps, run config
    else
        check user profile (default or specified via --profile)
        instantiate an aws client
        check valid credentials (via sts get caller identity)
        flag is_user
        get bucket name (from secret mgr)
    """


    def __init__(self, args):

        self.check_version()

        if args.command == 'config':
            success, msg = CmdConfig(args).run()
            print(msg)

        else:
            if profile_exists(args.profile):
                self.user_profile = get_profile(args.profile)
                self.aws = Aws(self.user_profile)

                if self.aws.is_valid_credentials():
                    # get bucket from local state if set
                    bucket = get_bucket()
                    if bucket:
                        self.aws.bucket_name = bucket
                    else:
                        try:
                            self.aws.get_bucket_name()
                        except:
                            print('Unable to get bucket')
                            sys.exit(1)

                    self.execute(args)

                else:
                    print('Invalid credentials')
                    sys.exit(1)

            else:
                print(f'Profile \'{args.profile}\' not found. Please run config command with your access keys')
                sys.exit(1)
                

    def check_version(self):

        today = date.today()
        last_checked = get_attr('version_checked')

        #print(f'today: {today}, last_checked: {last_checked}')

        if not last_checked or last_checked < today:

            resp = requests.get(f'https://pypi.org/pypi/{NAME}/json')
            latest_version = resp.json()['info']['version']

            if VERSION < latest_version:
                print(f'INFO: A new version of {NAME} is available. Run `pip install {NAME} --upgrade` to upgrade.')
            
            set_attr('version_checked', today)
    

    def execute(self, args):
        if args.command == 'create':
            success, msg = CmdCreate(self.aws, args).run()
            self.exit(success, msg)

        elif args.command == 'select':
            success, msg = CmdSelect(self.aws, args).run()
            self.exit(success, msg)

        elif args.command == 'list':
            success, msg = CmdList(self.aws, args).run()
            self.exit(success, msg)

        elif args.command == 'upload':
            success, msg = CmdUpload(self.aws, args).run()
            self.exit(success, msg)

        elif args.command == 'download':
            success, msg = CmdDownload(self.aws, args).run()
            self.exit(success, msg)

        elif args.command == 'delete':
            success, msg = CmdDelete(self.aws, args).run()
            self.exit(success, msg)

        elif args.command == 'sync':
            success, msg = CmdSync(self.aws, args).run()
            self.exit(success, msg)

    def exit(self, success, message):
        if message:
            print(message)

        if success:
            sys.exit(0)
        else:
            sys.exit(1)



import os
from tqdm import tqdm
from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool

from botocore.exceptions import ClientError
from boto3.s3.transfer import TransferConfig
from botocore.config import Config

from ait.commons.util.aws_client import Aws
from ait.commons.util.common import gen_uuid, format_err, INGEST_UPLOAD_AREA_PREFIX
from ait.commons.util.local_state import get_selected_area
from ait.commons.util.upload_service import notify_upload


class CmdSync:

    def __init__(self, aws: Aws, args):
        self.aws = aws
        self.args = args

    def run(self):
        if not self.aws:
            return False, 'You need configure your profile first'

        if self.aws.is_user:
            return False, 'You don\'t have permission to use this command'

        selected_area = get_selected_area()

        if not selected_area:
            return False, 'No area selected'
        
        dest_bucket, dest_env, dest_upload_area_uuid = self.args.INGEST_UPLOAD_AREA

        try:
            # Resources are not thread safe.
            # Low-level clients are thread safe. When using a low-level client, it is recommended to instantiate 
            # your client then pass that client object to each of your threads.

            s3_res = self.aws.common_session.resource('s3', config=Config(s3={'use_accelerate_endpoint': True}))
            s3_cli = s3_res.meta.client
            
            bucket = s3_res.Bucket(self.aws.bucket_name)

            fs = []
            total_size = 0

            # get all files from selected area
            for obj in bucket.objects.filter(Prefix=selected_area):
                # skip the top-level directory
                if obj.key == selected_area:
                    continue
                total_size += obj.size
                fs.append(obj)

            failed_fs = []
            
            def transfer(f):
                try:
                    
                    fname = f.key[37:]
                    content_type = ''
                    obj_ = s3_cli.head_object(Bucket=self.aws.bucket_name, Key=f.key)
                    if obj_ and obj_['ContentType']:
                        content_type = obj_['ContentType']
                        if "dcp-type=data" not in content_type:
                            content_type += '; dcp-type=data'

                    copy_source = {
                        'Bucket': self.aws.bucket_name,
                        'Key': f.key
                    }
                    dest_key = dest_upload_area_uuid + '/' + fname

                    s3_cli.copy(copy_source, dest_bucket, dest_key,
                                    Callback=pbar.update, 
                                    ExtraArgs={
                                        'ContentType': content_type,
                                        'MetadataDirective': 'REPLACE',
                                        },
                                    Config=get_transfer_config(f.size))

                    if not notify_upload(dest_env, dest_upload_area_uuid, fname):
                        failed_fs.append((f, 'Transferred. Notify failed.'))

                except ClientError as ex:
                    if ex.response['Error']['Code'] == 'NoSuchKey':
                        failed_fs.append((f, 'NoSuchKey'))
                    else:
                        failed_fs.append((f, str(thread_ex)))
                    pass

                except Exception as thread_ex:
                    failed_fs.append((f, str(thread_ex)))
                    pass

            print('Transferring...')
            pbar = tqdm(total=total_size, unit='B', unit_scale=True, desc=num_files(fs))
            pool = ThreadPool() # cpu_count() DEFAULT_THREAD_COUNT=25
            pool.map_async(transfer, fs)
            pool.close()
            pool.join()
            pbar.close()

            if failed_fs:
                print(f'{num_files(failed_fs)} failed to transfer: ')
                for f,err in failed_fs:
                    print(f'{f.key} {err}')
                return False, 'Transfer complete with error.'
            else:
                return True, 'Transfer complete.'
            
        except Exception as e:
            return False, format_err(e, 'sync')


def num_files(ls):
    l = len(ls)
    return f'{l} file{"s" if l > 1 else ""}'


# this is based on the dcplib s3_multipart module
KB = 1024
MB = KB * KB
MIN_CHUNK_SIZE = 64 * MB
MULTIPART_THRESHOLD = MIN_CHUNK_SIZE + 1
MAX_MULTIPART_COUNT = 10000 # s3 imposed


def get_transfer_config(filesize):
    return TransferConfig(multipart_threshold=MULTIPART_THRESHOLD, 
                          multipart_chunksize=get_chunk_size(filesize))


def get_chunk_size(filesize):
    if filesize <= MAX_MULTIPART_COUNT * MIN_CHUNK_SIZE:
        return MIN_CHUNK_SIZE
    else:
        div = filesize // MAX_MULTIPART_COUNT
        if div * MAX_MULTIPART_COUNT < filesize:
            div += 1
        return ((div + MB - 1) // MB) * MB



import json

from botocore.exceptions import ClientError

from ait.commons.util.command.area import CmdArea
from ait.commons.util.common import format_err
from ait.commons.util.local_state import get_selected_area

'''
ToDo
1. fix delete object
if an object key is specified e.g. abc.txt, current behaviour is deleting all objects with prefix 'abc.txt' for e.g. 'abc.txt2'

'''

class CmdDelete:
    """
    both admin and user, though user can't delete folder
    aws resource or client used in command - s3 resource (bucket.objects/ obj.delete)
    """

    def __init__(self, aws, args):
        self.aws = aws
        self.args = args

    def run(self):

        selected_area = get_selected_area()

        if not selected_area:
            return False, 'No area selected'

        try:
            if self.args.d:  # delete area
                if self.aws.is_user:
                    return False, 'You don\'t have permission to use this command'

                confirm = input(f'Confirm delete upload area {selected_area}? Y/y to proceed: ')

                if confirm.lower() == 'y':
                    print('Deleting...')

                    deleted_keys = self.delete_upload_area(selected_area, incl_selected_area=True)
                    for k in deleted_keys:
                        print(k)

                    # delete bucket policy for user-folder permissions
                    # only admin who has perms to set policy can do this
                    self.clear_area_perms_from_bucket_policy(selected_area)

                    # clear selected area
                    CmdArea.clear(False)
                return True, None

            if self.args.a:  # delete all files
                
                confirm = input(f'Confirm delete all contents from {selected_area}? Y/y to proceed: ')
                
                if confirm.lower() == 'y':
                    print('Deleting...')

                    deleted_keys = self.delete_upload_area(selected_area, incl_selected_area=False)
                    for k in deleted_keys:
                        print(k)

                return True, None

            if self.args.PATH:  # list of files and dirs to delete
                print('Deleting...')
                for p in self.args.PATH:
                    # you may have perm x but not d (to load or even do a head object)
                    # so use obj_exists

                    prefix = selected_area + p
                    keys = self.all_keys(prefix)

                    if keys:
                        for k in keys:
                            try:
                                self.delete_s3_object(k)
                                print(k + '  Done.')
                            except Exception as ex:
                                if 'AccessDenied' in str(ex):
                                    print('No permision to delete.')
                                else:
                                    print('Delete failed.')
                    else:
                        print(prefix + '  File not found.')
                return True, None
            else:
                return False, 'No path specified'

        except Exception as e:
            return False, format_err(e, 'delete')

    # based on obj_exists method
    def all_keys(self, prefix):
        keys = []
        response = self.aws.common_session.client('s3').list_objects_v2(
            Bucket=self.aws.bucket_name,
            Prefix=prefix,
        )
        for obj in response.get('Contents', []):
            keys.append(obj['Key'])
            
        return keys

    def delete_s3_object(self, key):
        s3_resource = self.aws.common_session.resource('s3')
        s3_obj = s3_resource.ObjectSummary(self.aws.bucket_name, key)
        s3_obj.delete()
        return key

    def delete_upload_area(self, selected_area, incl_selected_area=False):
        s3_resource = self.aws.common_session.resource('s3')
        bucket = s3_resource.Bucket(self.aws.bucket_name)
        deleted_keys = []
        objs_to_delete = bucket.objects.filter(Prefix=selected_area) if incl_selected_area else filter(lambda obj: obj.key != selected_area, bucket.objects.filter(Prefix=selected_area))
        for obj in objs_to_delete:
            obj.delete()
            deleted_keys.append(obj.key)

        return deleted_keys

    def clear_area_perms_from_bucket_policy(self, selected_area):
        s3_resource = self.aws.common_session.resource('s3')
        return CmdDelete.delete_dir_perms_from_bucket_policy(s3_resource, self.aws.bucket_name, selected_area)

    @staticmethod
    def delete_dir_perms_from_bucket_policy(s3_res, bucket_name, area_name):
        bucket_policy = s3_res.BucketPolicy(bucket_name)
        try:
            policy_str = bucket_policy.policy # throws NoSuchBucketPolicy
        except ClientError:
            policy_str = ''

        if policy_str:
            policy = json.loads(policy_str)
            policy_updated = False

            # remove any statement affecting single resource
            # (this also maintains backward compatibility with the previous way of adding
            # a statement per upload area)
            for stmt in policy['Statement']:

                if isinstance(stmt['Resource'], str) and area_name in stmt['Resource']:
                    policy_updated = True
                    policy['Statement'].remove(stmt) # cannot modify if removing item while iterating over list

            # now check statement with resource list
            for stmt in policy['Statement']:            
                if isinstance(stmt['Resource'], list):
                    # remove resource from resource list of statement but not statement
                    for res in stmt['Resource']:
                        if area_name in res:
                            policy_updated = True
                            stmt['Resource'].remove(res)

            if policy_updated:
                try:
                    if policy['Statement']:
                        bucket_policy.put(Policy=json.dumps(policy))  # throws MalformedPolicy (policy document exceeds the maximum allowed size of 20480 bytes)
                    else:
                        bucket_policy.delete()
                except ClientError:
                    pass



from ait.commons.util.common import serialize, deserialize
from ait.commons.util.settings import LOCAL_STATE_FILE


class LocalState:

    def __init__(self):
        self.bucket = None # default
        self.selected_area = None
        self.known_areas = []
        self.version_checked = None

    def select_area(self, area_name):
        self.selected_area = area_name

        if area_name not in self.known_areas:
            self.known_areas.append(area_name)

    def unselect_area(self):
        self.selected_area = None

    def __str__(self):
        if self.selected_area:
            s = f'Selected {self.selected_area}'
        else:
            s = 'No area selected'

        s += '\nKnown areas:\n'
        known_areas_minus_selected = [d for d in self.known_areas if d != self.selected_area]

        if known_areas_minus_selected:
            s += '\n'.join(map(str, known_areas_minus_selected))
        else:
            s += 'None'

        return s

def get_local_state():
    obj = deserialize(LOCAL_STATE_FILE)
    if obj and isinstance(obj, LocalState):
        return obj
    return LocalState()


def set_local_state(obj):
    serialize(LOCAL_STATE_FILE, obj)


def set_selected_area(area_name):
    set_attr('selected_area', area_name)


def set_bucket(bucket):
    set_attr('bucket', bucket)


def get_selected_area():
    return get_attr('selected_area')


def get_bucket():
    return get_attr('bucket')


def get_attr(name):
    obj = deserialize(LOCAL_STATE_FILE)
    if obj and isinstance(obj, LocalState):
        if hasattr(obj, name):
            return getattr(obj, name)
    return None


def set_attr(name, value):
    obj = deserialize(LOCAL_STATE_FILE)
    if obj is None or not isinstance(obj, LocalState):
        obj = LocalState()
    setattr(obj, name, value)
    serialize(LOCAL_STATE_FILE, obj)



import os

import botocore

from ait.commons.util.common import format_err
from ait.commons.util.file_transfer import FileTransfer, TransferProgress, transfer
from ait.commons.util.local_state import get_selected_area


class CmdDownload:
    """
    admin and user
    aws resource or client used in command - s3 resource (Bucket().upload_file)
    """

    def __init__(self, aws, args):
        self.aws = aws
        self.args = args
        self.files = []

    def run(self):

        selected_area = get_selected_area()

        if not selected_area:
            return False, 'No area selected'

        try:
            s3_resource = self.aws.common_session.resource('s3')
            bucket = s3_resource.Bucket(self.aws.bucket_name)

            # choice 1
            all_files = self.args.a  # optional bool

            fs = []
            if all_files:
                # download all files from selected area
                for obj in bucket.objects.filter(Prefix=selected_area):
                    # skip the top-level directory
                    if obj.key == selected_area:
                        continue
                    fs.append(FileTransfer(path=os.getcwd(), key=obj.key, size=obj.size))
            else:
                # choice 2
                # download specified file(s) only

                for f in self.args.f:
                    # check if f exists
                    key = f'{selected_area}{f}'
                    try:
                        # if you're able to download (s3:GetObject) you can do HEAD Object which
                        # is used by resource.ObjectSummary
                        obj_summary = s3_resource.ObjectSummary(self.aws.bucket_name, key)
                        obj_size = obj_summary.size
                    except botocore.exceptions.ClientError as e:
                        if e.response['Error']['Code'] == "404":
                            fs.append(FileTransfer(path=os.getcwd(), key=key, status='File not found.', complete=True))
                        elif e.response['Error']['Code'] == "403":
                            # An error occurred (403) when calling the HeadObject operation: Forbidden
                            fs.append(FileTransfer(path=os.getcwd(), key=key, status='Access denied.', complete=True))
                        else:
                            # Something else has gone wrong.
                            fs.append(FileTransfer(path=os.getcwd(), key=key, status='Download error.', complete=True))
                    else:
                        fs.append(FileTransfer(path=os.getcwd(), key=key, size=obj_size))

            def download(idx):
                try:
                    file = fs[idx].key
                    os.makedirs(os.path.dirname(file), exist_ok=True)

                    s3 = self.aws.new_session().resource('s3')
                    s3.Bucket(self.aws.bucket_name).download_file(file, file, Callback=TransferProgress(fs[idx]))

                    # if file size is 0, callback will likely never be called
                    # and complete will not change to True
                    # hack
                    if fs[idx].size == 0:
                        fs[idx].status = 'Empty file.'
                        fs[idx].complete = True
                        fs[idx].successful = True

                except Exception as thread_ex:
                    if 'Forbidden' in str(thread_ex) or 'AccessDenied' in str(thread_ex):
                        fs[idx].status = 'Access denied.'
                    else:
                        fs[idx].status = 'Download failed.'
                    fs[idx].complete = True
                    fs[idx].successful = False

            print('Downloading...')

            transfer(download, fs)

            self.files = [f for f in fs if f.successful]

            if all([f.successful for f in fs]):
                return True, 'Successful download.'
            else:
                return False, 'Failed download.'

        except Exception as e:
            return False, format_err(e, 'download')



from ait.commons.util.settings import AWS_ACCOUNT, IAM_USER

"""
User groups:
1. admin currently has full s3 access
2. other user has restricted access including (specified in user policy)

a) list upload area contents
    {
      "Sid": "AllowListBucketExceptRootDirectory",
      "Effect": "Allow",
      "Action": "s3:ListBucket",
      "Resource": "arn:aws:s3:::<bucket>",
      "Condition": {
        "StringLike": {
          "s3:prefix": [
            "*/*"
          ]
        }
      }
    }

b) default upload and delete (ux) permission on upload areas

    {
      "Sid": "DefaultUploadAreaAccess",
      "Effect": "Allow",
      "Action": [
        "s3:PutObject",
        "s3:DeleteObject",
        "s3:GetObjectTagging"
      ],
      "Resource": "arn:aws:s3:::<bucket>/*/*"
    }

    (and get obj tagging permission)


Upload area permissions
Abbrev  Desc.       S3 action           S3 operation/API request
                    (object ops perms)
u       upload      s3:PutObject        # PUT Object, POST Object, Initiate Multipart Upload, Upload Part, Complete Multipart Upload, PUT Object - Copy
d       download    s3:GetObject        # GET Object, HEAD Object, SELECT Object Content
x       delete      s3:DeleteObject     # DELETE Object

Possible combinations are 'u', 'ud', 'ux', 'udx'

"""

ALLOWED_PERMS = ['u', 'ud', 'ux', 'udx']
DEFAULT_PERMS = 'ux'

# constraints - in bucket policy
# ux       denyDelete                  -> u
# ux       allowDownload               -> udx
# ux       allowDownload + denyDelete  -> ud

def allowDownloadStmt():
    return {
    "Sid": "AllowDownload",
    "Effect": "Allow",
    "Action": "s3:GetObject",
    "Resource": [],
    "Principal": { "AWS": [f"arn:aws:iam::{AWS_ACCOUNT}:user/{IAM_USER}"] }
}

def denyDeleteStmt():
    return {
    "Sid": "DenyDelete",
    "Effect": "Deny",
    "Action": "s3:DeleteObject",
    "Resource": [],
    "Principal": { "AWS": [f"arn:aws:iam::{AWS_ACCOUNT}:user/{IAM_USER}"] }
}



from ait.commons.util.common import format_err, is_valid_area_name
from ait.commons.util.local_state import get_selected_area


class CmdList:
    """
    admin and user
    aws resource or client used in command - s3 resource (list_objects_v2, get_object_tagging)
    """

    def __init__(self, aws, args):
        self.aws = aws
        self.args = args
        
        self.s3_cli = self.aws.common_session.client('s3')

    def run(self):

        if self.args.b:  # list all areas in bucket
            if self.aws.is_user:
                return False, 'You don\'t have permission to use this command'
            
            try:
                folder_count = 0
                for area in self.list_bucket_areas():
                    k = area["key"]
                    self.print_area(k, area)
                    folder_count += 1
                print_count(folder_count)
                return True, None

            except Exception as e:
                return False, format_err(e, 'list')

        else:  # list selected area contents
            selected_area = get_selected_area()

            if not selected_area:
                return False, 'No area selected'

            try:
                selected_area += '' if selected_area.endswith('/') else '/'
                n, p = self.get_name_and_perms(selected_area)
                self.print_area(selected_area, dict(name=n, perms=p))

                file_count = 0
                for k in self.list_area_contents(selected_area):
                    print(k)
                    if not k.endswith('/'):
                        file_count += 1

                print_count(file_count)
                return True, None
            except Exception as e:
                return False, format_err(e, 'list')

    def print_area(self, k, area):
        print(k, end=' ')
        p = ''
        if 'perms' in area:
            p = area.get('perms') or ''
        print(p.ljust(3), end=' ')
        if 'name' in area:
            n = area.get('name')
            print(f'{n}' if n else '', end=' ')
        print()
    
    def get_name_and_perms(self, k):
        n, p = None, None
        try: 
            tagSet = self.s3_cli.get_object_tagging(Bucket=self.aws.bucket_name, Key=k)

            if tagSet and tagSet['TagSet']:
                kv = dict((tag['Key'], tag['Value']) for tag in tagSet['TagSet'])
                n = kv.get('name', None)
                p = kv.get('perms', None)
            else: # for backward compatibility get name and perms from metadata
                if not self.aws.is_user: # only admin can retrieve metadata (head_object)
                    resp = self.s3_cli.head_object(Bucket=self.aws.bucket_name, Key=k)
                    if resp and resp['Metadata']:
                        meta = resp['Metadata']
                        n = meta.get('name', None)
                        p = meta.get('perms', None)
        except:
            pass 
        return n, p

    def list_bucket_areas(self):
        areas = []
        result = self.s3_cli.list_objects_v2(Bucket=self.aws.bucket_name, Delimiter='/')
        dirs = result.get('CommonPrefixes', [])
        for d in dirs:
            k = d.get('Prefix')
            if is_valid_area_name(k):
                n, p = self.get_name_and_perms(k)
                areas.append(dict(key=k, name=n, perms=p))
        return areas

    def list_area_contents(self, selected_area):
        contents = []

        s3_resource = self.aws.common_session.resource('s3')
        bucket = s3_resource.Bucket(self.aws.bucket_name)

        for obj in bucket.objects.filter(Prefix=selected_area):
            k = obj.key
            if k != selected_area:
                contents.append(k)

        return contents


def print_count(count):
    if count == 0:
        print('No item')
    elif count == 1:
        print('1 item')
    else:
        print(f'{count} items')



import json
import requests
import urllib.parse


UPLOAD_SERVICE_API = 'https://upload__ENV__archive.data.humancellatlas.org'

"""
Upload service steps to trigger data validation
1. get creds from upload srv api
2. upload data via awscli or hca-util (set content-type accordingly)
3. post notification to upload srv api

API endpoints
POST /v1/area/{upload_area_uuid}/credentials    Create credentials for access to this upload area
POST /v1/area/{upload_area_uuid}/{filename}     Notify upload of uploaded file
HEAD /v1/area/{upload_area_uuid}                Check if an upload area exists
"""

def upload_api_url(env, upload_area_uuid):
    sub = '.' if env == 'prod' else f'.{env}.'
    url = UPLOAD_SERVICE_API.replace("__ENV__", sub) + '/v1/area/' + upload_area_uuid
    return url


def create_creds(env, upload_area_uuid):
    r = requests.post(upload_api_url(env, upload_area_uuid) + '/credentials')
    if r.status_code == 201:
        # Created
        return r.json()
    #elif r.status_code == 404:
    #    # Upload area not found
    #    print(r.json())


def notify_upload(env, upload_area_uuid, filename):
    try:
        notify_url = upload_api_url(env, upload_area_uuid) + '/' + urllib.parse.quote(filename)
        r = requests.post(notify_url)
        if r.status_code == 202:
            return True # File upload notification added to queue
    except requests.exceptions.RequestException as e:
        pass # silently pass
    return False # notify failed


def check_upload_area_exists(env, upload_area_uuid):
    try:
        r = requests.head(upload_api_url(env, upload_area_uuid))
        if r.status_code == 200:
            return True
        else:
            # Upload area does not exist.
            return False
    except requests.exceptions.RequestException as e:
        print(str(e))



import os
import configparser
from ait.commons.util.common import create_if_not_exists
from ait.commons.util.settings import AWS_CONFIG_FILE, AWS_CREDENTIALS_FILE, DEFAULT_REGION


class UserProfile:
    def __init__(self):
        self.access_key = None
        self.secret_key = None
        self.region = None

    def __repr__(self):
        return 'UserProfile()'

    def __str__(self):
        return f'UserProfile (access_key={self.access_key}, secret_key={self.secret_key}, region={self.region})'


def profile_exists(profile):
    # let's not bother checking CONFIG_FILE to see if region is set
    # we can always use default region
    if not os.path.exists(AWS_CREDENTIALS_FILE):
        return False

    credentials = configparser.ConfigParser(comment_prefixes='/', allow_no_value=True)
    credentials.read(AWS_CREDENTIALS_FILE)

    if credentials.has_section(profile):
        return True

    return False


def get_profile(profile):
    credentials = configparser.ConfigParser()
    credentials.read(AWS_CREDENTIALS_FILE)

    user_profile = UserProfile()

    if credentials.has_section(profile):
        user_profile.access_key = credentials[profile].get('aws_access_key_id')
        user_profile.secret_key = credentials[profile].get('aws_secret_access_key')

    config = configparser.ConfigParser()
    config.read(AWS_CONFIG_FILE)

    if config.has_section(f'profile {profile}'):
        user_profile.region = config[f'profile {profile}'].get('region')

    if not user_profile.region:
        user_profile.region = DEFAULT_REGION

    return user_profile


def set_profile(profile, region, access_key, secret_key):
    """.aws/config
    [profile {profile}]
    region = {region}
    """

    create_if_not_exists(AWS_CONFIG_FILE)

    # set comment_prefixes to a string which you will not use in the config file
    config = configparser.ConfigParser(comment_prefixes='/', allow_no_value=True)
    config.read(AWS_CONFIG_FILE)

    if not config.has_section(f'profile {profile}'):
        config.add_section(f'profile {profile}')
    config.set(f'profile {profile}', 'region', region)

    with open(AWS_CONFIG_FILE, 'w') as out:
        config.write(out)

    """.aws/credentials
    [{profile}]
    aws_access_key_id = {0}
    aws_secret_access_key = {1}
    """

    create_if_not_exists(AWS_CREDENTIALS_FILE)

    credentials = configparser.ConfigParser(comment_prefixes='/', allow_no_value=True)
    credentials.read(AWS_CREDENTIALS_FILE)

    if not credentials.has_section(f'{profile}'):
        credentials.add_section(f'{profile}')
    credentials.set(f'{profile}', 'aws_access_key_id', access_key)
    credentials.set(f'{profile}', 'aws_secret_access_key', secret_key)

    with open(AWS_CREDENTIALS_FILE, 'w') as out:
        credentials.write(out)

    print('Credentials saved.')



# common functions

import os
import pickle
import uuid

from ait.commons.util.settings import DEBUG_MODE, MAX_LEN_PROJECT_NAME

INGEST_UPLOAD_AREA_PREFIX = 's3://org-hca-data-archive-upload-'

def gen_uuid():
    return str(uuid.uuid4())


def is_valid_uuid(val):
    try:
        uuid.UUID(str(val))
        return True
    except ValueError:
        return False


def is_valid_project_name(name):
    return 0 < len(name) <= MAX_LEN_PROJECT_NAME


def is_valid_area_name(area_name):
    """Area name format: uuid with or without /
    """
    if len(area_name) < 36:
        return False
    elif len(area_name) == 36:  # uuid without /
        return is_valid_uuid(area_name)
    elif len(area_name) == 37 and area_name.endswith('/'):  # uuid with /
        uuid_part = area_name[0:36]
        return is_valid_uuid(uuid_part)
    return False


def serialize(name, obj):
    """Returns True if serialized."""
    try:
        pickle_out = open(name, 'wb')
        pickle.dump(obj, pickle_out)
        pickle_out.close()
        return True
    except Exception as e:
        print(str(e))
        return False


def deserialize(name):
    """Returns None if can't deserialize or not found."""
    try:
        pickle_in = open(name, 'rb')
        obj = pickle.load(pickle_in)
        pickle_in.close()
    except (OSError, IOError) as e:
        obj = None
    return obj


def create_if_not_exists(file):
    """
    Create the file if it does not exist
    :param file:
    :return:
    """
    if not os.path.exists(file):
        basedir = os.path.dirname(file)
        if not os.path.exists(basedir):
            os.makedirs(basedir)
        open(file, 'w').close()


def print_err(e, cmd):
    """
    Print user-friendly error message or exception details depending on DEBUG MODE
    :param e:
    :param cmd:
    :return:
    """
    print(format_err(e, cmd))


def format_err(e, cmd):
    """
    Format to user-friendly error message or exception details depending on DEBUG MODE
    :param e:
    :param cmd:
    :return:
    """
    err = str(e)

    if 'Forbidden' in err or 'AccessDenied' in err:
        'You don\'t have permission to use this command'

    if DEBUG_MODE:
        return f'An exception of type {e.__class__.__name__} occurred in command {cmd}.\nDetail: ' + str(e)
    else:
        return f'An error occurred in command: {cmd}'



from ait.commons.util.common import format_err
from ait.commons.util.local_state import get_selected_area, set_selected_area


class CmdSelect:
    """
    admin and user
    aws resource or client used in command - 
    """

    def __init__(self, aws, args):
        self.aws = aws
        self.args = args

    def run(self):
        try:
            if self.args.AREA:
                key = self.args.AREA if self.args.AREA.endswith('/') else f'{self.args.AREA}/'

                if self.aws.obj_exists(key):
                    set_selected_area(key)
                    return True, f'Selected upload area is {key}'
                else:
                    return False, "Upload area does not exist"
            else:
                selected_area = get_selected_area()
                if selected_area:
                    return True, 'Currently selected upload area is ' + get_selected_area()
                else:
                    return False, 'No upload area currently selected'

        except Exception as e:
            return False, format_err(e, 'select')



from pathlib import Path

NAME = 'covid-util'
VERSION = '0.0.4'
DESC = 'CLI tool for uploading data to the European COVID-19 data platform'
AUTHOR = 'hca-ingest-dev'
AUTHOR_EMAIL = 'hca-ingest-dev@ebi.ac.uk'

# when true, displays exception details; otherwise user-friendly error message
DEBUG_MODE = True

DIR_SUPPORT = False

MAX_DIR_DEPTH = 5

# user home directory
USER_HOME = str(Path.home())

# aws config and credentials files
AWS_CONFIG_FILE = USER_HOME + '/.aws/config'
AWS_CREDENTIALS_FILE = USER_HOME + '/.aws/credentials'

AWS_SECRET_NAME = 'covid/util/secret'

# default profile uses credentials from [DEFAULT_PROFILE] section of AWS_CREDENTIALS_FILE
# and config from [profile DEFAULT_PROFILE] section of AWS_CONFIG_FILE
DEFAULT_PROFILE = 'covid-util'
DEFAULT_REGION = 'us-east-1'

# local state for user
LOCAL_STATE_FILE = USER_HOME + '/.covid-util'

# 2 IAM user types: admin and (normal) user
# in the context of hca, admin=wrangler and user=contributor
IAM_USER = 'COVIDUser'
IAM_ADMIN = 'COVIDAdmin'

AWS_ACCOUNT = '871979166454'



COVID_UTIL = False
MAX_LEN_PROJECT_NAME = 36

if COVID_UTIL:
    # import covid_util specific settings
    from ait.commons.util.settings.covid_util import *
else:
    # default to hca_util settings
    from ait.commons.util.settings.hca_util import *
